# -*- coding: utf-8 -*-
"""
Created on Wed Dec 21 19:00:00 2022

@author: Anna Grim
@email: anna.grim@alleninstitute.org

"""

import json
import os
import shutil

import tensorstore as ts
import zarr
from tifffile import imread


# -- os utils ---
def listdir(path, ext=None):
    """
    Lists files in directory with "ext" if provided.

    Parameters
    ----------
    path : str
        Path to directory.
    ext : str, optional
        Extension of files of interest.

    Returns
    -------
    list[str]
        List of files in directory with "ext" if provided.
    """
    if ext is None:
        return [f for f in os.listdir(path)]
    else:
        return [f for f in os.listdir(path) if ext in f]


def mkdir(path):
    """
    Makes a directory at "path" if it does not already exist.

    Parameters
    ----------
    path : str
        Path to directory.

    Returns
    -------
    None.

    """
    if not os.path.exists(path):
        os.mkdir(path)


def rmdir(path):
    """
    Removes a directory at "path" if it already exist.

    Parameters
    ----------
    path : str
        Path to directory.

    Returns
    -------
    None.

    """
    if os.path.exists(path):
        shutil.rmtree(path)


# --- data structure utils ---
def check_edge(edge_list, edge):
    """
    Checks if "edge" is in "edge_list".

    Parameters
    ----------
    edge_list : list or set
        List or set of edges.
    edge : tuple
        Edge.

    Returns
    -------
    bool : bool
        Indication of whether "edge" is contained in "edge_list".

    """
    if edge in edge_list or (edge[1], edge[0]) in edge_list:
        return True
    else:
        return False


def remove_edge(edge_list, edge):
    """
    Checks whether "edge" is in "edge_list" and removes it.

    Parameters
    ----------
    edge_list : list or set
        List or set of edges.
    edge : tuple
        Edge.

    Returns
    -------
    edge_list : list or set
        Updated list or set of edges with "edge" removed if it was contained
        in "edge_list".

    """
    if edge in edge_list:
        edge_list.remove(edge)
    elif (edge[1], edge[0]) in edge_list:
        edge_list.remove((edge[1], edge[0]))
    return edge_list


# --- io utils ---
def read_tensorstore(path):
    """
    Reads neuroglancer_precomputed file at "path".

    Parameters
    ----------
    path : str
        Path to directory containing shard files.

    Returns
    -------
    ts.TensorStore
        Image volume.

    """
    dataset_ts = ts.open(
        {
            "driver": "neuroglancer_precomputed",
            "kvstore": {"driver": "file", "path": path},
        }
    ).result()
    return dataset_ts[ts.d["channel"][0]]


def read_n5(path):
    """
    Reads n5 file at "path".

    Parameters
    ----------
    path : str
        Path to n5.

    Returns
    -------
    np.array
        Image volume.
    """
    return zarr.open(zarr.N5FSStore(path), "r").volume


def read_tif(path):
    """
    Reads tif file at "path".

    Parameters
    ----------
    path : str
        Path to tif.

    Returns
    -------
    np.array
        Image volume.
    """
    return imread(path)


def write_txt(path, contents):
    """
    Writes "contents" to a .txt file at "path".

    Parameters
    ----------
    path : str
        Path that .txt file is written to.
    contents : list[str]
        Contents to be written to .txt file.

    Returns
    -------
    None

    """
    with open(path, "w") as file:
        for line in contents:
            file.write(line + "\n")


def write_json(path, contents):
    """
    Writes "contents" to a .json file at "path".

    Parameters
    ----------
    path : str
        Path that .txt file is written to.
    contents : dict
        Contents to be written to .txt file.

    Returns
    -------
    None

    """
    with open(path, "w") as f:
        json.dump(contents, f)



# -*- coding: utf-8 -*-
"""
Created on Wed Dec 21 19:00:00 2022

@author: Anna Grim
@email: anna.grim@alleninstitute.org

"""

from random import sample

import networkx as nx

import segmentation_skeleton_metrics.seg_metrics as sm
from segmentation_skeleton_metrics import nx_utils, utils


class MergeMetric(sm.SegmentationMetrics):
    """
    Class that evaluates the quality of a segmentation in terms of the
    number of merges.
    """

    def detect_mistakes(self):
        """
        Detects merges in the predicted segmentation.

        Parameters
        -------
        None

        Returns
        -------
        None

        """
        self.merged_edges = set()
        for graph in self.graphs:
            dfs_edges = list(nx.dfs_edges(graph))
            while len(dfs_edges) > 0:
                (i, j) = dfs_edges.pop(0)
                label_i = self.get_label(graph, i)
                label_j = self.get_label(graph, j)
                if super().is_mistake(label_i, label_j):
                    self.site_cnt += 1
                    super().log(graph, [(i, j)])
                    dfs_edges = self.explore_merge(
                        graph, dfs_edges, i, label_i
                    )
                    dfs_edges = self.explore_merge(
                        graph, dfs_edges, j, label_j
                    )
                elif label_i == 0:
                    dfs_edges = self.mistake_search(graph, dfs_edges, i)

        # Save Results
        super().write_results("merges")
        self.edge_cnt = len(self.merged_edges)

    def mistake_search(self, graph, dfs_edges, root):
        """
        Determines whether void region has a mistake.

        Parameters
        ----------
        graph : networkx.Graph
            Graph with potential split.
        dfs_edges : list[tuple]
            List of edges in order wrt a depth first search.
        root : int
            Edge where possible split starts.

        Returns
        -------
        list[tuple].
            Updated "dfs_edges".

        """
        # Search
        queue = [root]
        collisions = dict()
        visited = set()
        while len(queue) > 0:
            i = queue.pop(0)
            for j in nx_utils.get_nbs(graph, i):
                label_j = self.get_label(graph, j)
                if frozenset([i, j]) in visited:
                    continue
                elif label_j != 0 and label_j not in collisions.keys():
                    collisions[label_j] = j
                else:
                    queue.append(j)
                visited.add(frozenset([i, j]))
                dfs_edges = utils.remove_edge(dfs_edges, (i, j))

        # Check for split
        recorded = list()
        if len(collisions) > 1:
            root = sample(list(collisions.values()), 1)[0]
            for i in collisions.values():
                if i != root:
                    self.site_cnt += 1
                    recorded.append((root, i))

        if len(collisions) > 1:
            super().log(graph, recorded)
        return dfs_edges

    def explore_merge(self, graph, dfs_edges, root, root_label):
        """
        Traverses "graph" from "root" to determine how many and which
        edges are merged.

        Parameters
        ----------
        graph : networkx.Graph
            Graph that represents a neuron.
        dfs_edges : list[tuple]
            List of edges in graph ordered wrt a depth first search.
        root : int
            Root node of list of edges in "dfs_edges".
        val : int
            Value of root node.

        Returns
        -------
        list[tuple]
            List of merged edges.

        """
        visited = set()
        queue = [(-1, root)]  # parent, child
        while len(queue) > 0:
            # Visit
            i, j = queue.pop(0)
            label_j = self.get_label(graph, j)
            if label_j == root_label and i != -1:
                self.merged_edges.update(frozenset((i, j)))
                dfs_edges = utils.remove_edge(dfs_edges, (i, j))
            visited.add(j)

            # Populate queue
            for k in nx_utils.get_nbs(graph, j):
                condition1 = k not in visited
                condition2 = frozenset((j, k)) not in self.merged_edges
                if condition1 and condition2:
                    queue.append((j, k))

        return dfs_edges



# -*- coding: utf-8 -*-
"""
Created on Wed Dec 21 19:00:00 2022

@author: Anna Grim
@email: anna.grim@alleninstitute.org

"""

import networkx as nx
import random
import segmentation_skeleton_metrics.seg_metrics as sm
from segmentation_skeleton_metrics import nx_utils, utils


class SplitMetric(sm.SegmentationMetrics):
    """
    Class that detects splits in a predicted segmentation

    """

    def detect_mistakes(self):
        """
        Detects splits in the predicted segmentation.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        for graph in self.graphs:
            root = nx_utils.sample_node(graph)
            dfs_edges = list(nx.dfs_edges(graph, source=root))
            while len(dfs_edges) > 0:
                (i, j) = dfs_edges.pop(0)
                label_i = self.get_label(graph, i)
                label_j = self.get_label(graph, j)
                if super().is_mistake(label_i, label_j):
                    self.site_cnt += 1
                    super().log(graph, [(i, j)])
                elif label_j == 0:
                    dfs_edges = self.mistake_search(graph, dfs_edges, i, j)
        print("# Splits:", self.site_cnt)
        print("% Omit:", self.edge_cnt / self.count_edges())
        super().write_results("splits")

    def mistake_search(self, graph, dfs_edges, nb, root):
        """
        Determines whether complex mistake is a split.

        Parameters
        ----------
        graph : networkx.Graph
            Graph with possible split at "root".
        dfs_edges : list[tuple]
            List of edges to be processed for mistake detection.
        root : int
            Node where possible split starts.

        Returns
        -------
        list[tuple].
            Updated "dfs_edges" with visited edges removed.

        """
        # Search
        queue = [root]
        visited = set()
        collisions = dict()
        while len(queue) > 0:
            i = queue.pop(0)
            label_i = self.get_label(graph, i)
            visited.add(i)
            if label_i != 0:
                collisions[label_i] = i
            else:
                nbs =  nx_utils.get_nbs(graph, i)
                for j in [j for j in nbs if j not in visited]:
                    if utils.check_edge(dfs_edges, (i, j)):
                        queue.append(j)
                        dfs_edges = utils.remove_edge(dfs_edges, (i, j))
                    elif j == nb:
                        queue.append(j)

        self.edge_cnt += len(visited)

        # Check for split
        recorded = list()
        if len(collisions) > 1:
            k = random.sample(list(collisions.values()), 1)[0]
            for i in collisions.values():
                if i != k:
                    self.site_cnt += 1
                    recorded.append((k, i))
            super().log(graph, list(recorded))
        
        return dfs_edges
